<?php
/**
 * abook_imapfolder.php
 *
 * Backend for (personal) addressbook stored in special imap mails.
 *
 *  NOTE. This class should not be used directly. Use the
 *        "AddressBook" class instead.
 *
 * @package plugins
 * @subpackage abook_imapfolder
 * @author Samuel Graenacher <xplo@users.sourceforge.net>
 * @copyright Copyright &copy; 2005, Fr. Sauter AG
 * @license http://opensource.org/licenses/gpl-license.php GNU Public License v2
 */
if (!defined("SM_PATH"))
    define('SM_PATH','../../');
require_once(SM_PATH . 'functions/global.php');
require_once(SM_PATH . 'functions/i18n.php');
require_once(SM_PATH . 'functions/imap.php');
require_once(SM_PATH . 'functions/addressbook.php');
require_once(SM_PATH . 'functions/display_messages.php');
require_once(SM_PATH . 'class/deliver/Deliver.class.php');
require_once(SM_PATH . 'functions/mime.php');
include_once('var_export.php');

 
/*
 *********************************************************************
 *** Auxiliary functions
 * 
 * Mostly used for debug purposes "outside" of squirrelmail. 
 */
 /**
  * Function to emulate squirrel functions. (For external testing.)
  */
if (!function_exists('_')) {
    function _($t) {
        return $t;
    }
}
if (!class_exists('addressbook_backend')) {
    /**
     * Emulated class for non-squirrel testsetups.
     * @package abook_imapfolder
     */
    class addressbook_backend {
        function addressbook_backend() {
            // Does nothing..
        }
        
        function set_error($string) {
            $this->error .= "\n" . $string;
            return false;
        }
    }
}
/*
 **********************************************************************
 ** Main Class
 */
 
/**
 * Addressbook IMAP Storage Backend
 *
 * This class implements a squirrelmail addressbook backend using IMAP Folders
 * and messages as storage medium. Full read-write access is possible. Shared
 * folders and accessible folders from other users are prefixed, and are
 * writable aswell (if the IMAP ACL allows write access).
 *
 * General behaviour facts: 
 *  - Plugin looks at all mailboxes a user sees, and matches them against
 *    a certain name, to see wether it needs to consider them an addressbook
 *    folder.
 *  - Additionally it marks one (1) mailbox, beginning with 'INBOX' as the 
 *    user's personal addrmailbox.
 *  - The personal addrmailbox is the one the user sees with NO prefixes.
 *  - All other contacts from other mailboxes get a prefix in the form of
 *    "someprefix/nickname". (The way the prefixes are added is customizable
 *    inside the abook_imapfolder class.)
 *   The Addressbook Folders contain:
 *    - One email per contact. Subject is the name/nick of the contact, but can
 *      be completely arbitrary! (The code doesn't use it.)
 *    - Each email consists of machine-written but human-readable content.
 *    - The format is a simple but non-standard one. If there is already any rfc
 *      or standard for these kinda storage, the reader/writer classes are
 *      quickly adjusted.
 *
 * @package abook_imapfolder
 * @version 1.3
 * @author Samuel Graenacher
 * @copyright Fr. Sauter AG
 */
class abook_imapfolder extends addressbook_backend {
    /**
     * Backend type
     *
     * Can be 'local' or 'remote'
     * @var string backend type
     * @access public
     */
    var $btype = 'local';

    /**
     * Internal backend name
     * @var string
     * @access public
     */
    var $bname = 'imap';

    /**
     * Displayed backend name
     * @var string
     * @access public
     */
     var $sname;

    /**
     * Writeable flag
     * @var bool
     */
    var $writeable = true;

    /**
     * Specifies wether this plugin prefers to do the poutput sorting itself.
     *
     * @access public
     * @var bool
     */
    var $ownsort = false; // Indicate that we do sorting ourselves.
    
    /**#@+
     * @access private
     */

    /**
     * Turns on migration mode.
     *
     * Will be set to true if we should silently migrate existing entries.
     * @var bool
     */
    var $migrationmode = false;
    
    /**
     * Migration source backend.
     *
     * Where to migrate from. (Must be an enabled addressbook backend.)
     * @var string
     */
    var $migratefrom = 'local_file';
    
    /**
     * Will be set to true to speedup subsequent calls to the
     * migration functions.
     * @var bool
     */
    var $_migratedAlready = false;
    
    /**
     * The user's own folder (not a shared one).
     * Autogenerated later, DO NOT CHANGE OR SET IT HERE!
     * @var string
     */
    var $userFolder = ''; 
    
    /**
     * How to distinguish an addressbook folder from a normal one.
     * Folders ending with this string, are considered a contact box. It's
     * not a simple "ends in" comparison tho, the folder name (even of shared
     * folders must contain this.
     * @var string
     */
    var $folderTemplate = 'Contacts'; 
    
    /**
     * abookFolders starting with this, are considered "owned by user"
     * and end up being the default write target.
     * @var string
     */
    var $topLevelFolder = 'INBOX';
    
    /**
     * This array holds all Folders that contain Addressbook data
     * The UserFolder <i>should</i> be the first item (but check against 
     * the userFolder anyway).
     * @var array
     */
    var $interestingFolders = Array();
    
    /**
     * This array will hold <i>ALL</i> parsed and prepared addressbook data.
     * Try to keep copies of this at a minimum as it's prolly the biggest
     * var used over a long time.
     * @var array
     */
    var $addressStore = Array();
    
    /**
     * Unlike the $writable parameter below. this array holds a list of
     * the read/write permissions we have on the interestingFolders.
     * Do note, that this won't be filled before the folder have been examined.
     * @var array
     */
    var $writableFolder = Array();
    
    /**
     * This will contain the IMAP delimiter. Will be overwritten!
     * @var string
     */
    var $delimiter = '';
    
    /**#@+
    * These hold the possible IMAP namespaces a server can supply. And also
    * a seperate delimiter for each of the namespaces. (RFC allows that.)
    * @var string
    * @access private
    */
    var $nsPersonal = 'defnspersonal'; // During development 'INBOX.'
    var $nsdPersonal = '';
    var $nsOtherUsers = 'defnsusers'; // During development 'user.'
    var $nsdOtherUsers = '';
    var $nsSharedFolders = 'defnsshared'; // During development ''
    var $nsdSharedFolders = '';
    /**#@-*/
    
    /**
     * This char is used in nicknames to separate the true nick from the folder
     * name.
     * @var string
     */
    var $prefixDelimiter = '/';
    
    /**
     * This will be prepended to the prefix of contacts coming from other users.
     * 
     * Possible resulting addressbook entry: user/peterparker/salesxyz
     * @var string
     */
    var $prefixOtherUser = 'user/'; 

    /**
     * This will be prepended to the prefix of contacts coming from 
     * shared folders.
     * 
     * Possible resulting addressbook entry: shared/management/salesxyz
     * @var string
     */
     var $prefixSharedFolder = 'shared/';
    
    /**
     * Which keys can have more than one value?
     * @var array
     */
    var $multivalueKeys = Array('email');
    
    /**
     * Regex to use for splitting multival fields.
     * NOTE: The order in this array must be matching the one
     *       of $multivalueKeys !
     * @var array
     */
    var $multivalueSplitters = Array('/,\s*/');
    
    /**
     * String to use for joining multival fields.
     * NOTE: The order in this array must be matching the one
     *       of $multivalueKeys !
     * @var array
     */
    var $multivalueJoiners = Array(', ');
    
    /**
     * Array to determine which multival keys should be filtered for uniqueness
     * Again, same order as in $multivalueKeys !
     * @var array
     */
    var $multivalueUnique = Array(false);
    
    /**
     * Lists all elements that a search will look in.
     * @var array
     */
    var $searchableFields = Array('email', 'nick','givenname','surname',
                                  'name','label');
    
    
    /**
     * This holds the connection ID for our imap connection once we got it.
     * @var bool
     */
    var $imapconn = false;
    
    /**
     * This holds the debug string.
     * @var string
     */
    var $debug = '';
    
    /**
     * Debug Mode enabled or not?
     * @var bool
     */
    var $don = false;
    
    /**
     * Path to the encoder and decoder classes. It will first try and look
     * if "EntryReader" and "EntryWriter" are already defined classes. If not
     * It'll load "entryreader_class.php" and "entrywriter_class.php" from the
     * following folder.
     * @var string
     */
    var $pathToIOClasses = '';
    /*
    var $lang_prevdomain;
    var $lang_prevdomainpath;
    var $lang_owndomain = 'abook_imapfolder';
    var $lang_owndomainpath = '../../plugins/abook_imapfolder/locale';
    */
    
    /**
     * Internal optimization flag (to avoid redundand open() calls)
     * @var bool
     * @access private
     */
    var $opened = false;
    
    /**#@-*/
    /*
     **********************************************************************
     ** Methods
     */
    /**#@+
     * @access private
     */
    /**
     * Constructor
     *
     */
    function abook_imapfolder($param) {
        
        if (is_array($param) && isset($param['writeable'])) {
           $this->writeable = $param['writeable'];
        }

        // These vars will not be (optionally) overwritten by the 
        // values from the config.
        
        $this->loadConfig(array('folderTemplate','topLevelFolder',
                                'prefixDelimiter','prefixOtherUser',
                                'prefixSharedFolder','multivalKeys',
                                'multivalSplitters','mulltivalJoiners',
                                'multivalUnique','searchableFields',
                                'don','pathToIOClasses','migrationmode',
                                'migratefrom'));
                                
        $this->open(true);
        
        //$this->languageInit();
        $this->sname = $this->_("IMAP Contacts");
        //$this->languageClean();
    }
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    
    function _($s)
    {
        bindtextdomain('abook_imapfolder', SM_PATH . 
                       'plugins/abook_imapfolder/locale');
        textdomain('abook_imapfolder');
        $s = _($s);
        bindtextdomain('squirrelmail', SM_PATH . 'locale');
        textdomain('squirrelmail');
        return $s;
    }

    /*
    function languageInit() {
        if (textdomain(NULL) != $this->lang_owndomain) {
            $this->lang_prevdomain = textdomain(NULL);
            $this->lang_prevdomainpath = bindtextdomain(
                                            $this->lang_owndomain,
                                            $this->lang_owndomainpath
                                            );
            textdomain($this->lang_owndomain);
        }
    }
    
    function languageClean() {
        if (textdomain(NULL) == $this->lang_owndomain) {
            // Switch to the previous domain
            bindtextdomain($this->lang_prevdomain, $this->lang_prevdomainpath);
            textdomain($this->lang_prevdomain);
        }
    }
    
    function languageSetSMLang() {
        bindtextdomain('squirrelmail', '../../locale');
        textdomain('squirrelmail');
    }
    */
    
    /**
     *
     */
    function dbglog($text) {
        if ($this->don) {
            $this->debug .= $text;
            
            print "<pre>$text</pre>\n";
            global $abook_imapfolder_debug;
            $abook_imapfolder_debug = $this->debug;
        }
    }
    
    /**
     * This method loads the configuration file
     * and overlays the class variables specified as array elements
     * to this method.
     * @param array $varnames Classvariables to overlay.
     * @return bool
     */
    function loadConfig($varnames) {
        if (!is_array($varnames)) {
            return false;
        }
        global $cai;
        if (file_exists(SM_PATH . 'plugins/abook_imapfolder/config.php')) {
            include(SM_PATH . 'plugins/abook_imapfolder/config.php');
        } else {
            $this->dbglog('lC: Falling back to sample config.');
            include(SM_PATH . 'plugins/abook_imapfolder/config.php.sample');
        }
        reset($varnames);
        while(list($dummy,$cvarshort)=each($varnames)) {
            $cvar = 'cai_'.$cvarshort;
            if (!isset($cai["$cvarshort"])) {
                continue;
            }
            //eval("\$this->$cvarshort = \$cai[\"$cvarshort\"];");
            $thevar = $cvarshort;
            $this->$thevar = $cai["$cvarshort"];
            $this->dbglog("lC: $thevar is now set to \$this->".$thevar."\n");
        }
        return true;
    }
    
    /**
     * "(is)SET (and) NOT EMPTY"
     */
    function sne(&$somevar) {
        // Auxilliary function to save some space
        if (isset($somevar) && !empty($somevar)) {
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Micro log function used for debugging when output can't be given on
     * stdout.
     */
    function minilog($txt) {
        $filename = '/tmp/abook_imapfolder_debug.txt';
        $fd = fopen($filename, 'a+');
        if (!$fd) {
            return false;
        }
        $datepart = '['.date("Y-m-d H:i:s").'] ';
        $atxt = split("\n",$txt);
        if (is_array($atxt)) {
            reset($atxt);
            while (list($dummy, $line) = each($atxt)) {
                flock($fd, LOCK_EX);
                fwrite($fd, $datepart.$line."\n");
                flock($fd, LOCK_UN);
            }
        } else {
            fwrite($fd, $datepart.$txt."\n");
        }
        fflush($fd);
        fclose($fd);
    }
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    /**
     * This function looks for existing IMAP connections, both in SQWM
     * and in $this, and if it finds none, it registers a new one.
     * @return int
     */
    function getIMAPConn() {
        if ($this->imapconn === false) {
            // I hope type comparison works for socket checks

            // First we go and try to find any existing imap connection
            // created by squirrel.
            global $imapConnection;
            if (isset($imapConnection)&&($imapConnection != false)) {
                // Found one, using it.
                $this->imapconn = $imapConnection;
            } else {
                // Fone none, creating a new one.
                global $username, $imapServerAddress, $imapPort;
                sqgetGlobalVar('key', $key, SQ_COOKIE);
                global $onetimepad;
                sqgetGlobalVar('onetimepad', $onetimepad, SQ_SESSION);
                // ^-- is the line above really necessary?
                $this->imapconn = sqimap_login($username, $key, 
                                               $imapServerAddress, $imapPort, 
                                               0);
                // sqimap_login does its own exiting on errors
                // FIXME: Set last param to 1 on PROD!
            }
            // And fetching the delimiter if it isn't known already
            global $delimiter;
            if (!isset($delimiter) || empty($delimiter)) {
                $delimiter = sqimap_get_delimiter($this->imapconn);
            }
        } else {
            // Reuse existing connection if there's any..
            return $this->imapconn;
        }
    }
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    /**
     * This method populates the internal list of "interestingFolders" with
     * IMAP folders matching certain criterias. (See Class Properties)
     */
    function locateFolders() {
        $this->getIMAPConn();

        // Find the personal namespace, so we can make an exception for the
        // users own addressbook later on..
        $ft = $this->folderTemplate;
        if (sqimap_capability($this->imapconn, 'NAMESPACE')) {
            $read = sqimap_run_command($this->imapconn, 'NAMESPACE', true, $a, $b);
            unset($a); unset($b); 
            unset($regs);
            // New, untested namespace code follows...
            $bigassregexp = '#\* NAMESPACE\s+'.
                '(\(\("([^"]*)" "([^"])"\)(\(.*\))?\)|NIL)\s+'. 
                '(\(\("([^"]*)" "([^"])"\)(\(.*\))?\)|NIL)\s+'.
                '(\(\("([^"]*)" "([^"])"\)(\(.*\))?\)|NIL)#';
            // Following the RFC IMAP answers can be quite complex to parse
            // if a server really decides to be IMAP-feature-complete, but this
            // regex should cover fancy responses aswell.
            // regs[ 1] = don't use
            // regs[ 2] = personal namespace
            // regs[ 3] = personal delimiter
            // regs[ 4] = don't use (personal flags)
            // regs 5-8 repeat above regs for OTHER-USERS namespace
            // regs 9-12 repeat above regs for SHARED-FOLDERS namespace
            if (preg_match($bigassregexp,$read[0],$regs)) {
                $this->nsPersonal = $nsPersonal = $regs[2];
                $this->nsdPersonal = $this->delimiter = $nsDelimiter = $regs[3];
                $this->nsOtherUsers = $regs[6];
                $this->nsdOtherUsers = $regs[7];
                $this->nsSharedFolders = $regs[10];
                $this->nsdSharedFolders = $regs[11];
                
                $personalAddressbook = $this->nsPersonal.$ft;
            } else {
                // Garbage response (or regexp screwup)
                $personalAddressbook = $ft;
                $nsDelimiter = sqimap_get_delimiter($this->imapconn);
                $this->nsPersonal = $this->nsOtherUsers = 
                $this->nsSharedFolders = '';
                $this->nsdPersonal = $this->nsdOtherUsers =
                $this->nsdSharedFolders = $nsDelimiter;
            }
            /*
            // Old, working but non-rfc compatible namespace code
            if (preg_match('/\* NAMESPACE\s+\(+"([^"]+)"\s+"(.)"/',$read[0],$regs)) {
                $nsPersonal = $regs[1];
                $nsDelimiter = $regs[2];
                $personalAddressbook = $nsPersonal.$ft;
            } else {
                // Garbage response...
                $personalAddressbook = $ft;
                $nsDelimiter = sqimap_get_delimiter($this->imapconn);
            }
            */
        } else {
            // The server doesn't spew out a useful
            $personalAddressbook = $ft;
            $nsDelimiter = sqimap_get_delimiter($this->imapconn);
            $this->nsPersonal = $this->nsOtherUsers = 
            $this->nsSharedFolders = '';
            $this->nsdPersonal = $this->nsdOtherUsers =
            $this->nsdSharedFolders = $nsDelimiter;
        }
        $this->delimiter = $nsDelimiter;
        
        // Fetch all mailboxes a user sees..
        $allboxes = sqimap_mailbox_list_all($this->imapconn);
        $boxes = array();
        // Filter out the ones ending in our special _Addressbook keyword
        while(list($key,$box)=each($allboxes)) {
            if (preg_match('/(.*'.$ft.')\r?\n?$/',$box['unformatted'],$regs)) {
                // Found some folder that includes the folderTemplate
                // Analyzing now..
                
                if ($regs[1] == $personalAddressbook) {
                    // Found the user's personal Addressbook.
                    array_unshift($boxes, $regs[1]);
                    $this->userFolder = $regs[1];
                } elseif((strpos($regs[1],$this->nsPersonal) === 0) &&
                        (substr_count($regs[1],$this->nsdPersonal) >=2)) {
                    // Found some other subfolder below the personal Namespace
                    // ignore it..
                    continue; // Ignore it..
                } else {
                    $boxes[] = $regs[1];
                }
            } else {
                // Otherwise ignore->discard it
            }
        }
        unset($allboxes);
        $this->interestingFolders = $boxes;
    }
    
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    /**
     * This function queries the IMAP Folders to check if we have write
     * access on them.
     * @return bool
     */
    function checkFolderAccess($which='all') {
        $this->getIMAPConn();
        // Make sure we know about some folders...
        if (count($this->interestingFolders) <= 0) {
            if (!$this->locateFolders()) {
                return false;
            }
        }
        
        reset($this->interestingFolders);
        $wf = array();
        while (list($dummy,$mailbox)=each($this->interestingFolders)) {
            // Check if we need to be selective..
            if ($which != 'all') {
                if ($mailbox != $which) {
                    continue;
                }
            }
            // Checking mailbox..
            $mbinfo = sqimap_mailbox_select ($this->imapconn, $mailbox);
            if (strstr(strtolower($mbinfo['RIGHTS']), 'write')) {
                $wf["$mailbox"] = true;
            } else  {
                $wf["$mailbox"] = false;
            }
            // Cleaning up
            unset($mbinfo);
            sqimap_run_command($this->imapconn, "CLOSE",
                               true, $response, $message);
        }
        $this->writableFolder = $wf;
        unset($wf);
        if ($which != 'all') {
            return $this->writableFolder["$which"];
        }
        
        return true;
    }

    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///

    /**
     * Takes a folder and optionally an id of a mail to read into a string.
     * Depending on the $seperate param, it returns either a string or an array
     * of strings.
     * @return mixed
     * @param string $folder Name of the IMAP folder to process
     * @param string $which Which mail to read. (Got from {@link findMail})
     * @param bool $seperate Wether to group results into one big strings or an
     *                       array of small ones.
     */
    function readMailBodiesIntoString($folder,$which='all', $seperate = false) {

        $this->getIMAPConn();
        
        $mbinfo = sqimap_mailbox_select($this->imapconn, $folder);
        $mailsToProcess = $mbinfo['EXISTS'];
        if ($which=='all') {
            //$selection = '1:'.$mailsToProcess;
            $selection = '1:*';
        } else {
            $selection = $which.':'.$which;
        }
        if (strstr(strtolower($mbinfo['RIGHTS']), 'write')) {
            $this->writableFolder["$folder"] = true;
        } else {
            $this->writableFolder["$folder"] = false;
        }
        $fetch_ary = sqimap_run_command_list($this->imapconn, 
                                             'FETCH '.$selection.' '.
                                             '(BODY[TEXT])',
                                             false, $response, $message);
        
        // Compose these messages into one big parsable text string
        $skipLines = Array('/^\r?\n$/', '/^\s*\*.*FETCH/', '/^\)\r?\n$/');
        // NOTE!! The mails are not 1:1 as on the server!
        //        i.e. empty lines are skipped..
        if (!$seperate) {
            $rawData = '';
        } else {
            $rawData = Array();
        }
        for ($imail = 1; $imail <= $mailsToProcess; $imail++) {
            if ($seperate) {
                $rawData[$imail] = '';
            }
            $currentMail = $fetch_ary[($imail-1)];
            $lines = count($currentMail);
            for ($line = 0; $line < $lines; $line++) {
                $skipThisLine = false;
                reset($skipLines);
                while (list($dummy,$pattern)=each($skipLines)) {
                    if (preg_match($pattern, $currentMail[$line])) {
                        $skipThisLine = true;
                        break;
                    }
                }
                if ($skipThisLine == true) {
                    continue;
                }
                if ($seperate) {
                    $rawData[$imail] .= $currentMail[$line];
                } else {
                    $rawData .= $currentMail[$line];
                }
            }
        }
        $dummy = sqimap_run_command($this->imapconn, "CLOSE",
                                    true, $response, $message);
        return $rawData;
    }

    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    /**
     * Tries to locate a message in a given Folder based on a given
     * attrib/value pair. 
     *
     * NOTE: This does NOT use UID, even if the server supports em.
     *
     * @return int
     * @param string $folder Which folder to process.
     * @param string $attrib Name of the attrib to look for.
     * @param string $value Name of the value to look for in the given attrib.
     * @param int $limit Number of results to supply at most. (Default: inf)
     */
    function findMail($folder, $attrib, $value, $limit = -1) {
        
        $bodies = $this->readMailBodiesIntoString($folder,'all',true);
        if (!$bodies) { // Pass error thru
            return $bodies;
        }
        
        if ($this->don) {
            $this->dbglog("fM: bodies is ".var_export($bodies,1)."\n");
        }
        
        $bcount = count($bodies);
        $matchingIDs = Array();
        for ($mid = 1; $mid <= $bcount; $mid++) {
            if (!class_exists('EntryReader')) {
                require_once($this->pathToIOClasses.'entryreader_class.php');
            }
            $ereader = new EntryReader();
            $completenick = $this->folder2Nick($folder,'blah',$prefix);
            $ereader->setPrefixDelimiter($this->prefixDelimiter);
            $ereader->setPrefix('nick',$prefix);
            $ereader->readFromString($bodies[$mid]);
            // filter out the wanted attrib from the parsed data..
            reset($ereader->parsedData);
            $this->dbglog("fM: parsedData is: ".var_export($ereader->parsedData,1).
                "\n");
            if (!is_array($ereader->parsedData[0])) {
                continue;
            }
            while(list($key,$val)=each($ereader->parsedData[0])) {
                if ($key == $attrib) {
                    $this->dbglog("fM: val is: ".var_export($val,1)."\n");
                    if (!is_array($val)&&($val == $value)) {
                        // Matching mail found!
                        $matchingIDs[] = $mid;
                    } elseif(is_array($val)) {
                        reset($val);
                        foreach($val as $subval) {
                            $this->dbglog("fM: subvalcompare $subval against $value\n");
                            if ($subval == $value) {
                                $this->dbglog("fM: subvalmatch ".var_export($subval,1).
                                    "\n");
                                $matchingIDs[] = $mid;
                            }
                        }
                    } // Else.. no match
                }
            }
            // Cleanup
            unset($ereader);
        }
        
        // Strip out dupes caused by subvals
        $matchingIDs = array_unique($matchingIDs);
        $this->dbglog("fM: matchingIDs is: ".var_export($matchingIDs,1)."\n");
        reset($matchingIDs);
        if ($limit == -1) {
            return $matchingIDs;
        } elseif ($limit == 1){
            return array_shift($matchingIDs);
        } else {
            $runner = 0;
            $retvals = Array();
            while(count($matchingIDs)>0) {
                $runner += 1;
                $retvals[] = array_shift($matchingIDs);
                if ($runner >= $limit) {
                    break;
                }
            }
            return $retvals;
        }
        
        return false;
    }
    
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    
    /**
     * This method reads in all the data it can find and puts it into the
     * internal parsed storage ({@link $addressStore}).
     */
    function gatherAllData() {
        
        if ($this->delimiter == '') {
            $this->delimiter = sqimap_get_delimiter($this->imapconn);
        }
        $del = $this->delimiter;
        
        $this->locateFolders();
        if (count($this->interestingFolders) <= 0) {
            // Error: Didn't find any folder
            $this->dbglog('gAD: No interesting folders, exiting early.'."\n");
            return true;
            //return $this->set_error('gatherAllData: Didnt find any folder.');
        }
        
        $allParsed = Array(); // This will hold all parsed folder data
        $fc = count($this->interestingFolders);
        for ($i = 0; $i < $fc; $i++) {
            // Create parser
            if (!class_exists('EntryReader')) {
                require_once($this->pathToIOClasses.'entryreader_class.php');
            }
            $ereader = new EntryReader();
            // Build nick prefix..
            // FIXME: If this function works as expected, remove following lines
            /* * /
            // Redundant since translation funcs got introdced
            if ($this->interestingFolders[$fc] == $this->userFolder) {
                $prefix = '';
            } else {
                if (preg_match('/([^\\'.$del.']*)\\'.$del.
                               '?'.$this->folderTemplate.'$/',
                               $this->interestingFolders[$fc],
                               $regs)) {
                    // Check if prefix is another user or a shared folder
                    if (preg_match('/^'.preg_quote($this->nsOtherUsers).'/',
                                   $this->interestingFolders[$fc])) {
                        $prefix = $this->prefixOtherUser.$regs[1];
                    } else {
                        $prefix = $this->prefixSharedFolder.$regs[1];
                    }

                }
            }
            /* */
            $prefix = '';
            $completenick = $this->folder2Nick($this->interestingFolders[$i],
                                               'blah',
                                               $prefix);
            if (!$completenick) {
                return $this->set_error("gatherAllData: Couldn't resolve ".
                                 "nickprefix for folder ".
                                 $this->interestingFolders[$i].
                                 "\n");
            }
            
            $ereader->setPrefixDelimiter($this->prefixDelimiter);
            $ereader->setPrefix('nick',$prefix);
            
            // Read mailbox
            $rawString = $this->readMailBodiesIntoString(
                                     $this->interestingFolders[$i]);
            // Parse mails
            $ereader->readFromString($rawString);
            // Copy the parsed info in the collecting array
            if ((!is_array($ereader->parsedData))||
                (count($ereader->parsedData)<= 0)
                ) {
                continue;
            }
            
            // Add a group field to every contact.
            $parsedcount = count($ereader->parsedData);
            for($gi = 0; $gi <= $parsedcount; $gi++) {
                $ereader->parsedData[$gi]['group'][] = $prefix;
            }
            reset($ereader->parsedData);
            while(list($key,$val)=each($ereader->parsedData)) {
                $allParsed[] = $val;
            }
            // Cleanup
            unset($ereader);
        }
        
        $this->addressStore = $allParsed;
        unset($allParsed);
        
    }
    
    
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    // These two functions take care of the namespace
    // translations.
    /**
     * Converts a given foldername and contact name into our special
     * nickname format.
     *
     * USAGE EXAMPLE:
     *  (imap delimiter in examples assumed to be '.')
     *  IN:  'sharednamespace.partners._Addressbook'
     *  OUT: 'shared.sxx-somegroup:partners'
     *
     * This function can also be used to just get the nickprefix.
     * In that case, give any arbitrary contactname and only look
     * at the referenced $nickprefix as well as the return value
     * for error catching-
     *
     * @return string
     * @param string $folder IMAP Folder name.
     * @param string $contactname Name of the contact
     * @param string $nickprefix Will be filled with just the prefix part of the
     *                           new generated name.
     */
    function folder2Nick($folder, $contactname, &$nickprefix) {
        // STATUS: IMPLEMENTED
        // TESTS: NONE
        
        // Prepare some of the object vars
        $pou = preg_quote($this->prefixOtherUser);
        $psf = preg_quote($this->prefixSharedFolder);
        
        // We have three possible namespaces to catch
        // Is it a PERSONAL folder?
        if (preg_match('/^'.preg_quote($this->userFolder).'/',
                       $folder,$reg)) {
            // Looks personal
            $nick = $contactname;
            $nickprefix = '';
        } else {
            // Doesn't look personal
            // Now is it another users' addressbook or a shared folder?
            $rxu = '/^'.preg_quote($this->nsOtherUsers).
                   '([^'.preg_quote($this->nsdOtherUsers).']*).*'.
                   preg_quote($this->folderTemplate).'/';
            $rxs = '/^'.preg_quote($this->nsSharedFolders).
                  '([^'.preg_quote($this->nsdSharedFolders).']*).*'.
                   preg_quote($this->folderTemplate).'/';
            if (preg_match($rxu,$folder,$reg)) {
                // user.blah._Addressbook ->
                // user-blah:contactname
                $nick = $pou.$reg[1].
                        $this->prefixDelimiter.$contactname;
                $nickprefix = $pou.$reg[1];
                
            } elseif (preg_match($rxs,$folder,$reg)) {
                // somegroup._Addressbook ->
                // somegroup:contactname
                $nick = $psf.$reg[1].
                        $this->prefixDelimiter.$contactname;
                $nickprefix = $psf.$reg[1];
            } else {
                // UhOh, couldn't match the folder to any namespace..
                // this is BAD
                return $this->set_error('folder2Nick: Couldnt match "'.$folder.
                                 '" to any namescape!');
            }
        }
        
        return $nick;
    }
    
    /**
    * This method tries to find the individual parts (mainly the folder) of a
    * given encoded nick.
    * @param string $nick Encoded nickname (Example: folders/blah/someguy)
    * @param string $contactname Ref to a variable that will be the true nick
    * @return mixed Either the found folder or false.
    */
    function nick2Folder($nick, &$contactname) {
        // STATUS: NOTFULLY IMPLEMENTED (not working, too much copy/pasted)
        // TESTS: NONE
        // USAGE EXAMPLE:
        //  (imap delimiter in examples assumed to be '.')
        //  IN:  'shared.sxx-somegroup:partners'
        //  OUT: 'sharednamespace.partners._Addressbook'
        
        
        if (!strstr($nick, $this->prefixDelimiter)) {
            // Looks personal
            $contactname = $nick;
            $folder = $this->userFolder;
        } else {
            // Build the regexps for the matching..
            //$rxu = '/user.someguy._Addressbook/';
            //$rxs = '/somegroup._Addressbook/';
            $rxu = ''.preg_quote($this->prefixOtherUser).'([^'.
                   preg_quote($this->prefixDelimiter).']+)';
            $rxs = ''.preg_quote($this->prefixSharedFolder).'([^'.
                   preg_quote($this->prefixDelimiter).']+)';
            $this->dbglog("n2f: rxu is: $rxu\n");
            $this->dbglog("n2f: rxs is: $rxs\n");
            $allparts = explode($this->prefixDelimiter,$nick);
            $part2 = array_pop($allparts);
            $part1 = implode($this->prefixDelimiter, $allparts);
            $this->dbglog("n2f: part1 is: $part1\n");
            $this->dbglog("n2f: part2 is: $part2\n");
            
            if (preg_match($rxu,$part1,$regs)) {
                // Looks like a nick from another user..
                $this->dbglog("n2f: rxu matched. ".$regs[1]."\n");
                $contactname = $part2;
                $folder = $this->nsOtherUsers.$regs[1].$this->nsdOtherUsers.
                          $this->folderTemplate;
            } elseif (preg_match($rxs,$part1,$regs)) {
                // Looks like a nick from a shared folder..
                $this->dbglog("n2f: rxs matched. ".$regs[1]."\n");
                $contactname = $part2;
                $folder = $this->nsSharedFolders.$regs[1].
                          $this->nsdSharedFolders.$this->folderTemplate;
            } else {
                // Couldnt map... BAD
                $this->dbglog("n2f: Couldn't map nick $nick!\n");
                return $this->set_error("nick2Folder: Couldn't map nick $nick !\n");
            }
            
        }
        $this->dbglog("n2f: Returning folder $folder\n");
        return $folder;
    }
    
    /**
     * This method joins a given array using the correct glue string for the
     * given type.
     * @return string Joined array, ready for use in squirrel fields.
     * @param string $type To of values to join. (Example: 'email')
     * @param array $parsedarray
     */
    function multivalueJoin($type, $parsedarray) {
        if (!is_array($parsedarray)||!isset($parsedarray[0])) {
            return false; // Wrong format
        }
        $typeid = array_search($type, $this->multivalueKeys);
        if ($typeid === NULL) {
            // Not found in the multivalKeys.. treating as singlevalue.
            $this->dbglog('mvJ: singlevalue key ('.$parsedarray[0].')\n');
            return $parsedarray[0];
        }
        $squirrel = implode($this->multivalueJoiners[$typeid],
                            $parsedarray);
        $this->dbglog('mvJ: joined key ('.$squirrel.')\n');
        return $squirrel; 
    }
    /**
     * This method splits a given string using the correct splitter for the
     * given type.
     * @return array
     * @param string $type Type of value to split. (Example: 'email')
     * @param string $squirreldata
     */
    function multivalueSplit($type, $squirreldata) {
        $typeid = array_search($type, $this->multivalueKeys);
        if ($typeid === NULL) {
            // Not found in the multivalKeys.. treating as singlevalue.
            $this->dbglog('mvS: singlevalue key ('.$squirreldata.')\n');
            return array(0 => $squirreldata);
        }
        $multiar = preg_split($this->multivalueSplitters[$typeid],
                              $squirreldata);
                              
        if ($this->multivalueUnique[$typeid] == true) {
            $multiar = array_unique($multiar);
        }
        $this->dbglog('mvS: split key into '.count($multiar).' parts.\n');
        return $multiar; 
    }
    
    /**
     * Translates the keys and formats of the internal storage into the
     * format and names that squirrel expects.
     * @param array $parsed Datastructure in internal format.
     * @return array Datastructure in squirrel format.
     */
    function parsed2Squirrel($parsed) {
        $this->dbglog("Entered parsed2Squirrel().\n");
        if (!is_array($parsed)) {
            return $this->set_error('parsed2Squirrel: Not given an array!');
        }
        $res = Array();
        reset($parsed);
        while(list($dummy,$data) = each($parsed)) {
            $fdata = Array(); // Reset/Cleanup
            $this->dbglog("Entered mainloop in parsed2Squirrel().\n");
            // nickname
            if ((isset($data['nick'][0]))&&(!empty($data['nick'][0]))) {
                $fdata['nickname'] = $data['nick'][0];
            }
            // composed (full) name
            if (isset($data['surname'][0])&&
                !empty($data['surname'][0])) {
                
                if (isset($data['givenname'][0])&&
                    !empty($data['givenname'][0])) {
                    $preName = $data['givenname'][0].' ';
                } else {
                    $preName = '';
                }
                $fdata['name'] = $preName.$data['surname'][0];
            }
            // first-/givenname
            if ((isset($data['givenname'][0]))&&(!empty($data['givenname'][0])))
            {
                $fdata['firstname'] = $data['givenname'][0];
            }
            // lastname / name
            if ((isset($data['surname'][0]))&&(!empty($data['surname'][0]))) {
                $fdata['lastname'] = $data['surname'][0];
            } elseif ((isset($data['name'][0]))&&(!empty($data['name'][0]))) {
                // Found an (old) 'name' field... using this as lastname
                $fdata['lastname'] = $data['name'][0];
            } else {
                // Couldn't find ANY name.. exiting...
                if (isset($data['nick'])) {
                    $this->dbglog("Couldn't find any (sur)name for nick ".
                                    $data['nick']."\n");
                } else {
                    $this->dbglog("Couldn't find any (sur)name for unknown ".
                                    "nick.\n");
                }
                continue;
            }
            // email
            if ((isset($data['email'][0]))&&(!empty($data['email'][0]))) {
                $fdata['email'] = $this->multivalueJoin('email',$data['email']);
            }
            // label
            if ((isset($data['label'][0]))&&(!empty($data['label'][0]))) {
                $fdata['label'] = $data['label'][0];
            }
            // group
            if ((isset($data['group'][0]))&&(!empty($data['group'][0]))) {
                $fdata['group'] = $data['group'][0];
            }
            // Also add a few internal attributes
            $fdata['backend'] = $this->bnum;
            $fdata['source'] = &$this->sname;
            
            array_push($res, $fdata);
        }
        return $res;
    }
    
    /**
     * Translates the keys and formats of squirrel into the
     * internal format and names.
     * @param array $squirrel Datastructure in squirrel format.
     * @return array Datastructure in internal format.
     */
    function squirrel2parsed($squirrel) {
        $this->dbglog("Entered squirrel2Parsed().\n");
        if (!is_array($squirrel)) {
            return $this->set_error('squirrel2Parsed: Not given an array!');
        }
        $res = Array();
        reset($squirrel);
        
        if (!$this->sne($squirrel['nickname'])||
            !$this->sne($squirrel['email'])||
            !$this->sne($squirrel['lastname'])) 
        {
            return $this->set_error($this->_("Necessary fields missing."));
        }
        
        $res['nick'] = array(0 => $squirrel['nickname']);
        
        // We first try to parse the "name" field into first and last name..
        if ($this->sne($squirrel['name'])) {
            $nameparts = explode(' ',$squirrel['name']);
            // First get the surname...
            $res['surname'] = array(0 => array_pop($nameparts));
            $res['givenname'] = array(0 => implode(' ',$nameparts));
        }
        // ..but if they're set explicitely, we prefer them
        if ($this->sne($squirrel['firstname'])) {
            $res['givenname'] = array(0 => $squirrel['firstname']);
        }
        if ($this->sne($squirrel['lastname'])) {
            $res['surname'] = array(0 => $squirrel['lastname']);
        }
        
        if ($this->sne($squirrel['email'])) {
            $res['email'] = $this->multivalueSplit('email', $squirrel['email']);
        }
        
        if ($this->sne($squirrel['label'])) {
            $res['label'] = array(0 => $squirrel['label']);
        }
        
        
        return $res;
    }
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    /**
     * Physically (and logically) remove the given nick.
     *
     * This resolves the nick to a physical imap folder and mail, and nukes it.
     * @param string $rawnick Nickname to delete.
     * @return bool
     */
    function killNick($rawnick) {
        // STATUS: IMPLEMENTED
        // TESTS: NONE
        
        // This will remove the mail mathing a certain alias.
        // This is done by finding out the folder by splitting the prefix
        // off (if there is any). Then the mail id is located, flagged as
        // deleted and then expunged.
        
        $this->getIMAPConn();
        
        /* * /
        // Redundant since translation funcs got introdced
        // This commentsection can be removed once the thing is working...
        if (strstr($rawnick,$this->prefixDelimiter)) {
            // Remove in a shared folder
            list($folderpart,$nick) = explode($this->prefixDelimiter, $rawnick);
            //$folder = $folderpart.$this->delimiter.$this->folderTemplate;
            // Decide if prefix is another user or a shared folder..
            if (preg_match('/^'.$this->prefixOtherUsers.'(.*)/',
                           $folderpart,$reg)) {
                // nick is coming from another users main box
                $folder = $this->nsOtherUsers.$this->nsdOtherUsers.$reg[1].
                          $this->nsdOtherUsers.$this->folderTemplate;
            } else {
                $folder = $folderpart.$this->nsdSharedFolders.
                          $this->folderTemplate;
            }
        } else {
            // Remove in users folder.
            $nick = $rawnick;
            $folder = $this->userFolder;
        }
        /* */
        $nick = '';
        $folder = $this->nick2Folder($rawnick, $nick);
        $this->dbglog("kN: Set folder to $folder\n");
        if (!$folder) {
            return $this->set_error('killNick: Error translating folder.'."\n");
        }
        
        // Now open the folder, and get a full subjectlist from it
        $mbinfo = sqimap_mailbox_select($this->imapconn, $folder);
        $numMails = $mbinfo['EXISTS'];
        $subjectQuery = 
            'FETCH 1:* (BODY.PEEK[HEADER.FIELDS (Subject)])';
            // UID enabled: 'FETCH 1:* (BODY.PEEK[HEADER.FIELDS (Subject)])';
        $subjects = $read_list = Array();
        $read_list = sqimap_run_command_list($this->imapconn, $subjectQuery, 
                                             true, $response, $message, 
                                             true);
        /*
        $this->debug .="kN: dump of read_list: ".var_export($read_list,1)."\n";
        */
        foreach ($read_list as $r) {
            //$r = implode('',$r);
            /* first we unfold the header */
            $r = str_replace(array("\r\n", "\n\t","\n\s"),array("\n",'',''),$r);
            //if (!preg_match("/^\*\s+([0-9]+)\s+FETCH/iAU",$r[0], $regs)) {
            $regs = Array();
            if (preg_match("/^\*\s+([0-9]+)\s+FETCH/iAU",trim($r[0]), $regs)==0) {
                // Unknown response from server...
                $this->debug .="kN: Unknown response from server while".
                                 ' listing subjects'."\n";
                $this->debug .="kN: dump of r[0]: ".var_export($r[0],1)."\n";
                $this->debug .="kN: dump of regs: ".var_export($regs,1)."\n";
                return $this->set_error('kN: Unknown response from server while'.
                                 ' listing subjects'."\n");
            } else {
                $id = $regs[1];
            }
            $subject = $r[1];
            // Original squirrel code on the line below.. 
            //$subject = substr($subject,strlen($subject)+2);
            // ..and working code here..
            $dummy = explode(':', $subject, 2);
            $subject = trim($dummy[1]);
            unset($dummy);
            $subjects[$id] = $subject;
        }

        ///
        $res = $this->findMail($folder, 
                               'nick', 
                               $this->folder2Nick($folder, $nick, $dummy)
                               );
        $this->debug .="kn: findMail returned: ".var_export($res,1)."\n";
        ///
        
        reset($res);
        $delcount = 0;
        while(list($dummy,$idToDel) = each($res)) {
            $this->dbglog("Deleting from $folder mailid $idToDel.\n");
            global $uid_support;
            $uidtemp = $uid_support; // Save server uid capability
            $uid_support = false; // Force server uid cap to off
            $mbinfo = sqimap_mailbox_select($this->imapconn, $folder);
            // Do the ACTUAL delete!
            sqimap_messages_flag($this->imapconn, $idToDel, $idToDel, 'Deleted',
                                 true);
            $delcount += sqimap_mailbox_expunge($this->imapconn,
                                                $folder,
                                                true);
            $uid_support = $uidtemp; // Restore uid cap
            unset($uidtemp);
        }
        
        /*
        $this->debug .="kN: found ".count($subjects)." mails in this folder.\n";
        $this->debug .="kN: dump of subjects: ".var_export($subjects,1)."\n";
        // Compare each subject against $nick
        reset($subjects);
        $delcount = 0;
        while(list($id,$sub)=each($subjects)) {
            $this->dbglog("kN: Comparing sub: $sub to nick: $nick\n";
            if ($sub == $nick) {
                $this->dbglog("kN: Compare MATCHED! Found the contact mail!".
                                "(Id: $id)\n";
                // Kill all exact matches. (Must be exact so we don't
                // accidentally wipe contacts due to short new contact names!)
                // Also note that we use a lowlevel func here because the
                // highlvl squirrel one would use the trash.
                /* * /
                // Variant 1 using some squirrel functions...
                global $uid_support;
                $uidtemp = $uid_support; // Save server uid capability
                $uid_support = false; // Force server uid cap to off
                $mbinfo = sqimap_mailbox_select($this->imapconn, $folder);
                sqimap_messages_flag($this->imapconn, $id, $id, 'Deleted',true);
                $delcount += sqimap_mailbox_expunge($this->imapconn,
                                                    $folder,
                                                    true);
                $uid_support = $uidtemp; // Restore uid cap
                unset($uidtemp);
                /* */
                /* * /
                // Variant 2 using other squirrel function and a small 
                // workaround
                global $move_to_trash;
                $tmpmvtrash = $move_to_trash; // Save preference
                $move_to_trash = false; // Overwrite preference
                sqimap_mailbox_select($this->imapconn, $folder);
                sqimap_messages_delete($this->imapconn, $id, $id, true);
                $delcount += sqimap_mailbox_expunge($this->imapconn, $folder, 
                                                    true);
                $move_to_trash = $tmpmvtrash; // Restore preference
                unset($tmpmvtrash); // Cleanup
                /* */
                
                /* * /
                // Variant 3, the do-it-yourself one (bad..)
                global $uid_support;
                sqimap_mailbox_select($this->imapconn, $folder);
                $readstore = sqimap_run_command($this->imapconn, 
                                'STORE '.$id.':'.$id.' +FLAGS (\Deleted)', true,
                                $response, $message, $uid_support);
                $this->dbglog("kN: readstore is: ".var_export($readstore,1).
                                "\n";
                $readexp = sqimap_run_command_list($this->imapconn, 
                                'EXPUNGE', true,
                               $response, $message, false);
                $this->dbglog("kN: expunge response is: $response\n";
                $this->dbglog("kN: expunge message  is: $message\n";
                $cnt = 0;
                if (is_array($readexp)) {
                    foreach ($readexp as $rx) {
                        $this->dbglog("kN: readexp line: $rx\n";
                        if (preg_match('/^\*\s+[0-9]+\sEXPUNGE/AUi',$rx,
                                       $dummy)
                            ) {
                            $cnt++;
                        }
                    }
                }
                $delcount += $cnt;
                /* * /
            }
        }
        */
        // Explicitely expunge the marked messages now.
        
        return $delcount;
        //return 1; //DBGREMOVE
    }

    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    /**
     * This physically adds a new contact. It is given an array ins squirrel
     * format containing the nick and the other required info and will create
     * the physical storage mail for it.
     *
     * IDEAS:
     *  - if contact goes in personal box, check if that box already exists
     *    using sqimap_mailbox_exists,
     *  - if it doesn't, set a sane userFolder assumption
     *    persnamespace+folderTemplate and try to create it
     *    using sqimap_mailbox_create followed by an sqimap_unsubscribe
     * @param array $contact All contact info in squirrel format.
     * @return bool
     */
    function createContact($contact) {
        
        $contactName = '';
        $writeIntoFolder = $this->nick2Folder($contact['nick'][0], 
                                              $contactName);
        $subject = $contactName;
        
        ///////
        // Folder checks..
        //  personal?
        if ($this->userFolder == $writeIntoFolder) {
            // yup goes into personal..
            // but do we even have a personal folder?
            if ($this->userFolder == '') {
                // nope, need to create a new one..
                $this->userFolder = $this->nsPersonal.$this->folderTemplate;
                $writeIntoFolder = $this->userFolder;
            }
            // does it exist?
            if (!sqimap_mailbox_exists($this->imapconn, $writeIntoFolder)) {
                // doesn't exist, trying to create it
                sqimap_mailbox_create($this->imapconn, $writeIntoFolder, '');
                if (!sqimap_mailbox_exists($this->imapconn, $writeIntoFolder)) {
                    // creation wasn't successful, erroring out..
                    return $this->set_error( 
                        sprintf(
                            $this->_("Error creating personal Addressbox %s"),
                            $writeIntoFolder
                            )
                        );
                }
                sqimap_unsubscribe($this->imapconn, $writeIntoFolder);
            }
            // now it does exist..
            
        } else {
            // Nope, not personal...
            // does it exist?
            if (!sqimap_mailbox_exists($this->imapconn, $writeIntoFolder)) {
                // doesn't exist, trying to create it
                sqimap_mailbox_create($this->imapconn, $writeIntoFolder, '');
                if (!sqimap_mailbox_exists($this->imapconn, $writeIntoFolder)) {
                    // creation wasn't successful, erroring out..
                    return $this->set_error(
                        sprintf(
                            $this->_("Error creating personal Addressbox %s"),
                            $writeIntoFolder
                            )
                        );
                }
                sqimap_unsubscribe($this->imapconn, $writeIntoFolder);
            }
            // now it does exist..
        }
        // Reload everything we know about folders
        $this->locateFolders();
        if (!$this->checkFolderAccess($writeIntoFolder)) {
            return $this->set_error(
                $this->_("You do not have the rights to change this Addressbox."
                )); 
        }
        // End of Folder checks
        
        if (!class_exists('EntryWriter')) {
            require_once($this->pathToIOClasses.'entrywriter_class.php');
        }
        $ewriter = new EntryWriter();
        $ewriter->setPrefixDelimiter($this->prefixDelimiter);
        $this->dbglog("cC: given contact is: ".var_export($contact,1)."\n");
        $body = $ewriter->writeSingleEntry($contact);
        $this->dbglog("*** ewriter debug start ***\n");
        $this->dbglog($ewriter->errors .= "\n");
        $this->dbglog("*** ewriter debug stop ***\n");
        unset($ewriter); // Free up some space
        
        $entryMessage = new Message();
        $rfc822_header = new Rfc822Header();
        $entryMessage->rfc822_header = $rfc822_header;
        $entryMessage->reply_rfc822_header = '';
        $rfc822_header = $entryMessage->rfc822_header;
        $from_addr = '"Addressbook" <addressbook@abook-imap.squirrel>';
        $rfc822_header->from = $rfc822_header->parseAddress($from_addr,true);
        $rfc822_header->to = $rfc822_header->parseAddress($from_addr,true);
        $rfc822_header->subject = $subject;
        $content_type = new ContentType('text/plain');
        $rfc822_header->encoding = '8bit';
        //$content_type->properties['charset']='ASCII';
        $this->dbglog("cC: Trying to add this body:\n$body\n");
        $entryMessage->setBody($body);
        
        $rfc822_header->content_type = $content_type;
        $entryMessage->rfc822_header = $rfc822_header;
        require_once(SM_PATH . 'class/deliver/Deliver_IMAP.class.php');
        $imap_deliver = new Deliver_IMAP();
        $length = $imap_deliver->mail($entryMessage);
        sqimap_append ($this->imapconn, $writeIntoFolder,
                       $length);
        $imap_deliver->mail($entryMessage, $this->imapconn);
        sqimap_append_done ($this->imapconn, $writeIntoFolder);
        unset($imap_deliver);
        unset($entryMessage);
        sqimap_mailbox_select ($this->imapconn, $writeIntoFolder);
        
        // This is really nasty: The sq functions to append/create imap
        // mails don't provide any success checks. :(
        return true;
    }
    
    /**
     * This can be called by methods that need an up to date internal storage.
     * @param bool $new Completely ignored.
     * @return bool
     * @deprecated
     */
    function open($new = false) {
        if ($this->opened) {
            return true;
        }
        
        // Create an IMAP connection if we don't have any..
        if ($this->imapconn === false) { // Yes, do a typecheck
            $this->getIMAPConn();
        }
        
        if (count($this->addressStore) == 0) {
            $gadret = $this->gatherAllData();
            if (!$gadret) {
                return $gadret;
            }
        }
        
        $this->opened = true;
        return true;
    }
    
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///

    /**
     * This method controls the migration.
     */
    function doMigration() {
        // No point in running this multiple times.
        if ($this->_migratedAlready==true) {
            $this->dbglog('dM: Already migrated.'."\n");
            return true;
        }
        $this->dbglog('dM: started doMigration().'."\n");
        $this->_migratedAlready = true; // We set this prevent rec calling.
        if ($this->migrationmode == true) {
            $this->dbglog('dM: Migration starting..'."\n");
            // Migration mode on. 
            $migbook = addressbook_init(true, true);
            if (!$migbook) {
                // Addressbook initialising failed.
                $this->dbglog('dM: abook initialisation failed.'."\n");
            } else {
                // Finding the source backend num
                $cbs = count($migbook->backends);
                for ($cb = 0; $cb < $cbs; $cb++) {
                    if ($migbook->backends[$cb]->bname == 
                        $this->migratefrom) {
                        $sourcenum = $cb;
                        $this->dbglog('dM: found migration source: '.$cb."\n");
                        break;
                    }
                }
                // Found the source number into $sourcenum
                $this->dbglog('dM: Going into migrateFrom().'."\n");
                //$this->minilog($this->debug); $this->debug = '';
                $this->migrateFrom($sourcenum, $migbook);
                $this->dbglog('dM: Got out of migrateFrom().'."\n");
                //$this->minilog($this->debug); $this->debug = '';
            }
        } else {
            $this->dbglog('dM: Migrationmode not enabled.'."\n");
        }
        if (isset($migbook)) {
            $migbook = NULL;
            unset($migbook);
        }
        //echo "<pre>".$this->debug."</pre><br/>\n";
        //$this->debug = ''; // Avoid duplicate debuglog printing.
    }

    /**
     * This method does the actual movement of elements from one
     * backend to the next.
     *
     * IDEA: This takes a source backend (given by a fullblown
     *       Addressbook instance and the number of the backend,
     *       reads in all entries (list_addr), puts it into $this
     *       and then removes all of em in the source backend.
     * @param int $snum
     * @param Addressbook $book
     */
    function migrateFrom($snum, &$book) {
        if (!$book) {
            return false;
        }
        
        $srclist = $book->list_addr($snum);
        if (!$srclist) {
            $this->dbglog('mF: Bad list_addr result.'."\n");
            return $srclist;
        }
        if (count($srclist)== 0) {
            // Nothing to convert.
            $this->dbglog('mF: Nothing to migrate.'."\n");
            return true;
        }
        reset($srclist);
        while(list($dummy,$row) = each($srclist)) {
            unset($addret);
            
            if (!isset($row['nickname'])||empty($row['nickname'])) {
                continue; // Skip corrupt entries
            }
            
            // Check if the nick exists already. Don't migrate it if if so.
            $this->dbglog('mF: Looking up '.$row['nickname']."\n");
            $exists = $this->lookup($row['nickname']);
            if (is_array($exists) && (count($exists) > 0)) {
                $this->dbglog('mF: Didn\'t migrate nick ('.
                                $row['nickname'].'). Exists already.'.
                                "\n");
                $this->dbglog('mF: $exists is: '.var_export($exists,1)."\n");
                continue;
            }
            $this->dbglog('mF: Adding '.$row['nickname']."\n");
            
            // Sanity check on incomplete entries
            if (!isset($row['email'])||empty($row['email'])) {
                $row['email'] = '?';
            }
            if (!isset($row['lastname'])||empty($row['lastname'])) {
                $row['lastname'] = '?';
            }
            
            // Last minute cleanup of "bad" values
            // We clone the array so we can still reference to the original
            // values.
            $row2add = $row;
            $row2add['nickname'] = str_replace('/','',$row2add['nickname']);
            
            $addret = $this->add($row2add);
            if ($addret != false) {
                $remret = $book->remove($row['nickname'], $snum);
                if (!$remret) {
                    $this->dbglog('mF: Error removing migrated nick ('.
                                  $row['nickname'].")\n");
                }
                
            } else {
                $this->dbglog('mF: Error adding contact to us.'."\n");
                $this->dbglog('mF: Error was: '.$this->error."\n");
            }
        }
        
    }
    ////////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    ///
    
    
    /**
     * Deprecated function. Avoid using it, becuae it doesn't do anything
     * useful.
     * @deprecated
     */
    function close() {
        return;
    }
    
    /**
     * Converts a GLOB expression into a perl regex.
     * @param string $glob The GLOB Expression (Example: 'blah*')
     * @param bool $escape Wether the output should be escaped already.
     * @param bool $exact Wether or not there may be anything in front or 
     *                    behind the given glob.
     * @return string
     */
    function glob2regexp($glob, $escape = true, $exact = false) {
        $reg = $glob;
        
        // Escape if requested
        if ($escape) {
            $reg = preg_quote($reg);
            // Translating escaped glob symbols to nonescaped regex ones
            $reg = str_replace('\*','.*',$reg);
            $reg = str_replace('\?','.?',$reg);
        } else {
            // Translating glob symbols to regex ones
            $reg = str_replace('*','.*',$reg);
            $reg = str_replace('?','.?',$reg);
        }
        
        if ($exact == true) {
            $reg = '^'.$reg.'$';
        }
        // Adding the rest of necessary regex markers
        $reg = ':('.$reg.'):i';
        return $reg;
    }

    /**
     * Searches the given (internal) datastructure in the given fields
     * for the given pattern.
     * @param array $parsedData
     * @param array $searchFields
     * @param string pattern
     * @return mixed
     */
    function searchParsedData($parsedData, $searchFields, $pattern) {
        if ((!is_array($parsedData))||(!is_array($searchFields))) {
            // Unusable input format
            return false;
        }
        
        $returnSet = Array();
        $matchFound = false;
        reset($parsedData);
        // Loop thru each potential parsed object
        while(list($key,$fields) = each($parsedData)) {
            $matchFound = false;
            // $key = object number (?) FIXME: verify
            // $fields = array of all fields and their contents
            
            // Loop thru each field 
            while(list($fieldname,$fieldcontents) = each($fields)) {
                // Check if we need to consider this attribute
                if (in_array(strtolower($fieldname),$searchFields)) {
                    // Need to consider this indeed..
                    
                    if (!is_array($fieldcontents)) {
                        // Entry has wrong format, skipping.
                        continue;
                    }
                    // Loop thru all subattributes, usually only one.
                    $countSubContents = count($fieldcontents);
                    for ($si = 0; $si < $countSubContents; $si++) {
                        $this->dbglog("sPD: Looking for pattern: $pattern\n");
                        $this->dbglog("sPD: I see a fieldtype $fieldname");
                        $this->dbglog("sPD: set to: ".$fieldcontents[$si]."\n");
                        if (preg_match($pattern,$fieldcontents[$si])) {
                            $matchFound = true;
                            $this->dbglog("searchParsedData: MATCH!\n");
                            break;
                        }
                    }
                    if ($matchFound) { // Match found, skip the useless rest
                        break;
                    }
                }
                if ($matchFound) { // Match found, skip the useless rest
                    break;
                }
            }
            
            if ($matchFound) {
                array_push($returnSet, $fields);
                $matchFound = false;
            }
        }
        
        $this->dbglog("returnSet is: ".var_export($returnSet,1)."\n");
        return $returnSet;
    }
    
    ////////////////////////////////////////////////////////////////
    /// Sorting functions //////////////////////////////////////////
    ////////////////////////////////////////////////////////////////
    ///
    
    /**
     * Callback function used to sort nicknames.
     * 
     * Sorts while being aware of the different sources/folders that contacts
     * can come from.
     * @param array $a
     * @param array $b
     */
    function squirrelSorter($a, $b) {
        global $srtdbg; // Nasty workaround the fact that callbacks don't
                        // seem to see their parent object?!
        
        // First we do some weighting and see if we can do sort by different
        // weights.
        
        $weights = Array(
            1 => preg_quote($this->prefixSharedFolder), // Sort bottom-most
            2 => preg_quote($this->prefixOtherUser), // Sort in the middle
            3 => ".*" // Sort highest, fallback.
            );
            
        for ($i = 1; $i <= count($weights); $i++) {
            if (preg_match(':^'.$weights[$i].':', $a['nickname'])) {
                $srtdbg .= "squirrelSorter: nick ".$a['nickname'].
                    " has weight: $i pattern was: ".':^'.$weights[$i].':'."\n";
                $weightA = $i;
                break;
            }
        }
        for ($i = 1; $i <= count($weights); $i++) {
            if (preg_match(':^'.$weights[$i].':', $b['nickname'])) {
                $srtdbg .= "squirrelSorter: nick ".$b['nickname'].
                     " has weight: $i pattern was: ".':^'.$weights[$i].':'."\n";
                $weightB = $i;
                break;
            }
        }
        
        // Easiest scenario: the two got different weights => return the one
        // with the higher weight.
        if ($weightA != $weightB) {
            if ($weightA > $weightB) {
                $srtdbg .= "squirrelSorter: weight check won: ".$a['nickname'].
                           "\n";
                return -1;
            } else {
                $srtdbg .= "squirrelSorter: weight check won: ".$b['nickname'].
                           "\n";
                return 1;
            }
        }
        
        // When we reach this point, we're comparing two items with equal
        // weight.
        $srtdbg .= "squirrelSorter: same weight nicks: ".$a['nickname'].
             " and ".$b['nickname']."\n";
        $nres = strcasecmp($a['lastname'], $b['lastname']);
        if ($nres < 0) {
            $srtdbg .= "squirrelSorter: lastname: ".$a['lastname'].
             " has won over ".$b['lastname']."\n";
        } elseif ($nres > 0) {
            $srtdbg .= "squirrelSorter: lastname: ".$b['lastname'].
             " has won over ".$a['lastname']."\n";
        } else {
            $srtdbg .= "squirrelSorter: lastname: ".$b['lastname'].
             " is identical to ".$a['lastname']."\n";
        }
        return $nres;
        
        
    }
    
    /**
     * Sorts the given resultset.
     * @param array $results Resultsets in internal format.
     * @param array $options Unused.
     * @return array
     */
    function sortSquirrelResults($results, $options =Array()) {
        global $srtdbg;
        $srtdbg = '';
        $processed = $results;
        $sorted = uasort($processed, array($this,'squirrelSorter'));
        /* * /
        echo "<pre>squirrelSorter produced this debug output:\n".
            $srtdbg."</pre>\n";
        /* */
        $srtdbg='';
        /* * /
        echo "list_addr: before sortSquirrelResults: <pre>".
        var_export($results,1)."</pre>\n";
        echo "<pre>=====================================\n</pre>";
        echo "list_addr: sortSquirrelResults processed: <pre>".
        var_export($processed,1)."</pre>\n";
        /* */
        
        // Renumbering/Rekeying the array, cause sort doesn't touch the keys.
        $pcount = count($processed);
        $linedUp = Array();
        for ($i = 0; $i < $pcount; $i++) {
            $processed[$i]['sorted'] = true;
            $linedUp[] = array_shift($processed);
        }
        /* * /
        echo "<pre>=====================================\n</pre>";
        echo "list_addr: sortSquirrelResults lined up: <pre>".
        var_export($linedUp,1)."</pre>\n";
        /* */
        
        return $linedUp;
    }
    
    ///
    /////////////////////////////////////////////////////////////////
    /**#@-*/
    /* ========================== Public ======================== */
     
    /**
     * Searches the entire backend for the given glob expression.
     * @param string $expr
     * @return array Results in squirrel array format.
     */
    function &search($expr) {
        $ret = array();
        if(!$this->open()) {
            $this->dbglog('search: error on open()'."\n");
            $this->doMigration();
            $this->set_error('Error initializing required objects.');
            return array();
        }
        $this->doMigration();
         
        // To be replaced by advanded search expression parsing
        if (is_array($expr)) {
            return;
        }

        // Make regexp from glob'ed expression
        $expr = $this->glob2regexp($expr);

        $fieldsToSearch = $this->searchableFields;
        
        if (count($this->addressStore) <= 0) { 
            // No addresses found yet? Try a search again...
            $gadret = $this->gatherAllData();
            if (!$gadret) {
                return $gadret;
            }
        }
        
        $foundData = $this->searchParsedData($this->addressStore, 
                                             $fieldsToSearch, $expr);
        if ((!$foundData) || (count($foundData)<=0)) {
            return false;
        }

        $squirrelResults = Array();
        reset($foundData);
        $squirrelResults = $this->parsed2Squirrel($foundData);
        
        return $squirrelResults;
    }
    
    /**
     * Looks up all available info on a single given nickname
     *
     *  IDEAS:
     * This could be optimized to only read in the requested
     * contact mail/folder if it can guess it, but as a first
     * implementation this will simply gatherAllData() and filter
     * all for the wanted contact..
     *
     * @param string $alias Nickname to retrieve info for.
     * @return array Infoset in squirrel format.
     */
    function &lookup($alias) {
        if (empty($alias)) {
            return array();
        }
         
        $alias = strtolower($alias);
        
        if (!$this->open()) {
            $this->dbglog('lookup: error on open()'."\n");
            $this->doMigration();
            $this->set_error('Error initializing required objects.');
            return array();
        }
        
        $this->doMigration();
        
        $expr = $this->glob2regexp($alias, true, true);
        
        
        $fieldsToSearch = Array('nick');
        // ^FIXME: Make above line a class property
        
        if (count($this->addressStore) <= 0) { 
            // No addresses found yet? Try a search again...
            $gadret = $this->gatherAllData();
            if (!$gadret) {
                return $gadret;
            }
        }
        if (count($this->addressStore) <= 0) { 
            // Still no addresses found? Erroring out...
            $gadret = $this->gatherAllData();
            if (!$gadret) {
                return $gadret;
            }
        }
        
        // echo "<pre>Address store is: ".var_export($this->addressStore,1).
             // "\n</pre>";
        
        $foundData = $this->searchParsedData($this->addressStore, 
                                             $fieldsToSearch, $expr);
        if ((!$foundData) || (count($foundData)<=0)) {
            return false;
        }
        $squirrelResults = Array();
        reset($foundData);
        $squirrelResults = $this->parsed2Squirrel($foundData);
        
        // This function may not return more than one result!
        return $squirrelResults[0]; 
    }

    /**
     * Returns all addresses the backend knows of in squirrel format.
     * @return array
     */
    function &list_addr() {
        $this->dbglog("Entered list_addr().\n");
        $ret = array();
        if (!$this->open()) {
            $this->dbglog('list_addr: error on open()'."\n");
            $this->doMigration();
            $this->set_error('Error initializing required objects.');
            return array();
        }
        $this->doMigration();
        
        
        $this->dbglog("list_addr: survived open().\n");
        
        $squirrelResults = $this->parsed2Squirrel($this->addressStore);
        
        $sortedSquirrelResults = $this->sortSquirrelResults($squirrelResults);
        
        return $sortedSquirrelResults;
    }

    /**
     * Adds a new contact.
     * Public squirrelmail interface function.
     *
     * This method rearranges $userdata and the given
     * attribute-names and then invokes our {@link createContact}()
     * method.
     * @param array $userdata
     * @return bool
     */
    function add($userdata) {
        if (!$this->open()) {
            /*
            return $this->set_error(
                $this->_("Error opening addressbook module.")
            );*/
        }

        if (!$this->writeable) {
            return $this->set_error($this->_("Addressbook is read-only"));
        }
        
        $this->dbglog("add: Got this userdata: ".var_export($userdata,1)."\n");
        
        $ud = &$userdata;
        // Check for the absolute minimum of necessary fields
        if (!isset($ud['nickname']) || empty($ud['nickname']) ||
            !isset($ud['email'])    || empty($ud['email'])    ||
            !isset($ud['lastname']) || empty($ud['lastname'])) {
            return $this->set_error($this->_("Necessary fields missing."));
        }
        
        /*
        // This comment block has been deprecated.. remove before publish..
        // Translating squirrel fields to abook_imapfolder ones.
        $id = Array();
        $id['nick'] = $ud['nickname'];
        $id['email'] = $ud['email'];
        if (isset($ud['firstname'])) {
            $id['givenname'] = $ud['firstname'];
        }
        $id['surname'] = $ud['lastname'];
        if (isset($ud['label'])) {
            $id['label'] = $ud['label'];
        }
        */
        $id = $this->squirrel2parsed($ud);
        
        $ret = $this->createContact($id);
        /*
        if ($this->don) {
            print "<pre>".$this->debug."</pre>";
        }
        */
        // Doing a data refresh to pick up created ones right away.
        $this->addressStore = array();
        $this->gatherAllData();
        
        if (!$ret) {
            return $this->set_error($this->_("Error in createContact"));
        } else {
            return true;
        }
    }

    /**
     * Removes a given nick/alias from the storage completely.
     * 
     * This function handles single and multiple deletes transparently
     * @param mixed $alias Nickname(s) to remove.
     * @return bool
     */
    function remove($alias) {
        $this->dbglog("remove: Got Param ".var_export($alias,1)."\n");
        // Handle multiple deletions by recursing ourselves.
        if (is_array($alias)) {
            foreach($alias as $ca) {
                if (!is_string($ca) || empty($ca)) { // Minimal sanity check
                    return $this->set_error(
                        $this->_("Delete input in wrong format.")
                    );
                }
                $recret = $this->remove($ca);
                if ($recret == false) { // Had an error apparently
                    return $recret;
                }
            }
        } else {
            
            $ret = $this->killNick($alias);
            if ($this->don) {
                print "<pre>".$this->debug."</pre>";
            }
            // Refresh
            $this->addressStore = Array();
            $this->gatherAllData();
            
            if ($ret > 0) {
                return true;
            } else {
                return $this->set_error(
                    sprintf($this->_("Could not remove alias '%s'"), $alias)
                    );
            }
        }
    }

    /* Modify address */
    /**
     * Overwrites an existing contact with the new given userdata.
     * @param string $alias Nickname of the contact to modify.
     * @param array $userdata New (and complete) new info.
     * @return bool
     */
    function modify($alias, $userdata) {
        if (!$this->open()) {
            return $this->set_error('Error initializing required objects.');
            //return true;
        }

        $contactname = '';
        $this->debug = '';
        $folder = $this->nick2Folder($alias, $contactname);
        if ($this->don) {
            echo "<pre>".$this->debug."</pre>";
        }
        if (!$this->checkFolderAccess($folder)) {
            return $this->set_error(sprintf(
                  $this->_("You cannot write in the Addressbook folder '%s'!"),
                  $folder));
        }
        
        /* See if user exist */
        $ret = $this->lookup($alias);
        if (empty($ret)) {
            return $this->set_error(sprintf(
                                        $this->_("User '%s' does not exist"),
                                        $alias)
                                    );
        }
        
        // Sanity check BEFORE we remove stuff.. yes duplicated code but
        // necessary to do it here!
        $ud = &$userdata;
        // Check for the absolute minimum of necessary fields
        if (!isset($ud['nickname']) || empty($ud['nickname']) ||
            !isset($ud['email'])    || empty($ud['email'])    ||
            !isset($ud['lastname']) || empty($ud['lastname'])) {
            return $this->set_error($this->_("Necessary fields missing."));
        }
        
        $ret = $this->remove($alias);
        if ((!$ret)||(empty($ret))) {
            return $this->set_error(sprintf(
                                        $this->_("Could not remove alias '%s'"),
                                        $alias)
                                    );
        }
        
        $ret = $this->add($userdata);
        if ((!$ret)||(empty($ret))) {
            return $this->set_error(sprintf(
                                        $this->_("Could not add alias '%s'"),
                                        $alias)
                                    );
        } else {
            return true;
        }
        
    }
} /* End of class abook_database */

?>
<?php /* 
jEdit local settings 
:indentSize=4:noTabs=true:maxLineLen=80:wrap=hard:mode=php:
*/ ?>
